#ifndef COMMON_H
#define COMMON_H

#include "fstream"
#ifdef OPENCV_2_3_1
#include <opencv2/opencv.hpp>
#include <opencv2/legacy/legacy.hpp>
#include <opencv2/legacy/compat.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#else
#include "cv.h"
#include "cxcore.h"
#include "highgui.h"
#endif
#include "time.h"
#include <stdio.h>
#include <stdlib.h>
#ifdef _WIN32
#include "conio.h"
#endif
typedef unsigned int TTT;
//0 kmeans, 1 em, 2 hierarchical, 3 sparse histogram
#define CLUSTERING_TYPE 0

#define max(x,y) ((x) > (y) ? (x) : (y))
#define min(x,y) ((x) < (y) ? (x) : (y))
#define isNan(x) ((x) != (x) ? 1 : 0)
using namespace std;
#define SHOW_IMAGE 0

template<class T>
inline T sqr(T x) { return x*x; };

class Method{
public:	
	int dim;
	int clusterNum;
	char name[20];
	bool nowTesting;
	virtual void extractFeatures(float* X, int* nrKeypoints, IplImage* img) = 0;	
};

class MethodRepo{
	std::map<std::string, Method *> repo;
public:
	static MethodRepo *instance() {		
		static MethodRepo *inst = new MethodRepo();
		return inst;
	}
	void add(std::string name, Method *m) {
		repo[name] = m;
	}
	Method *get(std::string name) {
		return repo[name];
	}
};

struct histbin{
	TTT i;
	float x;
	histbin(TTT i,float x):i(i),x(x){};
	histbin(){histbin(0,0);};
	void operator = (const histbin* q){
		i = q->i;
		x = q->x;
	}
	void set(TTT ii, float xx){
		i=ii;
		x=xx;
	}
};

//parameter handling -- attribute value pair structure
struct av{
	char attr[256];
	float value;

	av(char* at, float v):value(v){
		strcmp(attr,at);
	};

	av():value(0){	};

	void set(char* at, float v){
		strcpy(attr,at);
		value = v;
	};
};

//structure to hold parameters
struct params{
	av avs[15];
	int mnr;//number of methods
	int attrNr;
	int trainingIndex;
	char methodName[5][256];

	params(){
		attrNr=0;
	}	

	void setAtr(char* attr, float value){
		int i;
		for(i=0;i<attrNr;i++)
			if ( strcmp(avs[i].attr, attr) == 0 )
			{
				avs[i].value = value;
				break;
			}

		if(i==attrNr)
		{
			attrNr++;			
			avs[i].set(attr, value);
		}
	}

	float get(char* attr){
		for(int i=0;i<attrNr;i++)
			if ( strcmp(avs[i].attr, attr) == 0 )
				return avs[i].value;
		return 0;
	}	
};

struct ErrorItem{
	float err;
	int imgnum;
	ErrorItem(){err = 1e35f; imgnum = -1;};
	ErrorItem(float e, int num):err(e),imgnum(num){};
};

class ErrorItemComp{
	public:
    bool operator()(ErrorItem& t1, ErrorItem& t2)
    {
       if (t1.err > t2.err) return true;      
       return false;
    }
};

struct ErrorItemL{
	float err;
	unsigned long long imgnum;
	ErrorItemL(){err = 1e35f; imgnum = 0;};
	ErrorItemL(float e, unsigned long long num):err(e),imgnum(num){};
};

class ErrorItemCompL{
	public:
    bool operator()(ErrorItemL& t1, ErrorItemL& t2)
    {
       if (t1.err > t2.err) return true;      
       return false;
    }
};


/*
//file reader from folder for windows
class FileGetter{
	WIN32_FIND_DATAA found;	
	HANDLE hfind;
	char folder[255];			
	int chk;
	bool first;

public:
	FileGetter(char* folderin,char* ext){		
		strcpy(folder,folderin);
		char folderstar[255];	
		if( !ext ) strcpy(ext,"*");
		sprintf(folderstar,"%s\\*.%s",folder,ext);
		hfind = FindFirstFileA(folderstar,&found);
		first = 1;
		//skip .
		//FindNextFileA(hfind,&found);		
	}

	int getNextFile(char* fname){
		//skips .. when called for the first time
		if( first )
		{
			strcpy(fname, found.cFileName);				
			first = 0;
			return 1;
		}
		else{
			chk=FindNextFileA(hfind,&found);
			if (chk)
				strcpy(fname, found.cFileName);				
			return chk;
		}
	}

	int getNextAbsFile(char* fname){
		//skips .. when called for the first time
		if( first )
		{
			sprintf(fname, "%s\\%s", folder, found.cFileName);			
			first = 0;
			return 1;
		}
		else{
			chk=FindNextFileA(hfind,&found);
			if (chk)
				sprintf(fname, "%s\\%s", folder, found.cFileName);				
			return chk;
		}
	}
};
//*/

#endif