#include "common.h"
#include "kmeansBased.h"
#include "imgOperations.h"
#include "corel5k.h"

extern params param;
extern char method_names[5][256];
//#include "descr.h"
#ifdef _WIN32
#include "direct.h"
#include <windows.h>
#include <conio.h>
#endif
#include <stdio.h>
#include <time.h>
#include <queue>
using namespace std;

//dir points to the database root folder
Corel5k::Corel5k(const char* dir){
	strcpy(baseDir,dir);		
	maxImgSize = 120;

	hist = new float[classNr];
	hists = new float[trainImgNr*20];
	imgnums = new int[trainImgNr+1];
	frec = new int[classNr];
	wh = new int[classNr];
	wordcounts = new int[trainImgNr];
	float* weights;
	cooc = new float[classNr*classNr];
#if  CLUSTERING_TYPE == 0
	//k-means
	centers = new cv::Mat[param.mnr*trainImgNr];	
#elif CLUSTERING_TYPE == 2
	//em
	models = new gmm[param.mnr*trainImgNr];
#elif CLUSTERING_TYPE == 3
	shists.resize(trainImgNr);
#endif
}

//move all annotation files to the temporary folder datai
void Corel5k::clean(int i){
#ifdef _WIN32
	char str[256];
	sprintf(str,"%s%s\\data%02d",baseDir,method_names[0],i);
	mkdir(str);
	sprintf(str," move /y %s%s\\*.* %s%s\\data%02d",baseDir,method_names[0],baseDir,method_names[0],i);
	printf("%s\n",str);
	system(str);
#endif
}

//read meta-data from data folder
void Corel5k::readData(){
	char fname[255];
	sprintf(fname,"%sdata\\image_nums",baseDir);
	FILE *f = fopen(fname,"r");
	int imgnum=-1;
	int nr=0;
	do{
		if( fscanf(f,"%d",&imgnum) < 1 )
			break;
		imgnums[nr]=imgnum;
		nr++;				
	}while(imgnum>0);
	fclose(f);

	sprintf(fname,"%sdata\\document_words",baseDir);
	f = fopen(fname,"r");
	nr = 0;
	for(int i=0;i<classNr;i++)
		frec[i]=0;
	for(int i=0;i<trainImgNr;i++){
		for(int j=0;j<5;j++){
			fscanf(f,"%d",&ground_truth[i][j]);
		if( ground_truth[i][j] > 0)
			frec[ground_truth[i][j]]++;
		}
	};
	fclose(f);

	sprintf(fname,"%sdata\\words",baseDir);
	f = fopen(fname,"r");
	nr = 0;
	do{
		nr++;
		if( fscanf(f,"%s",words[nr]) < 1 )
			break;				
	}while(1);
	fclose(f);

	sprintf(fname,"%sdata\\test_1_document_words",baseDir);
	f = fopen(fname,"r");	
	memset(wh,0,classNr);
	for(int i=0;i<testImgNr;i++)
	{
		for(int j=0;j<5;j++)
		{
			if( fscanf(f,"%d",&test_gt[i][j]) < 1 )
				goto break1;	
			if (test_gt[i][j])
				wh[test_gt[i][j]]++;		
		}
	};
break1:
	fclose(f);

	sprintf(fname,"%sdata\\word_counts",baseDir);
	f = fopen(fname,"r");	
	for(int i=0;i<trainImgNr;i++)
		fscanf(f,"%d",&wordcounts[i]);
	fclose(f);
	
	if(1)
		cooccurrence();
	sprintf(fname,"%sdata\\cooc",baseDir);
	cooc = new float[classNr*classNr];
	f = fopen(fname,"r");	
	for(int i=1;i<classNr;i++)
		for(int j=1;j<classNr;j++)
			fscanf(f,"%f",cooc+i*classNr+j);
	fclose(f);
}

//save descriptor X to txt file
void Corel5k::saveDescr(float *X, int imgNum, float* hist){
	char outFname[255];
	sprintf(outFname,"%s%s\\%d\\%d.txt",baseDir,methodName,imgNum-imgNum%1000,imgNum);
	FILE* f = fopen(outFname,"w");
	fprintf(f,"%d %d\n",clusterNum,dim);
	for(int nr=0;nr<clusterNum;nr++){
		for(int d=0;d<dim;d++){
			if( !isNan(X[nr*dim+d]) )
				fprintf(f,"%f ",X[nr*dim+d]);
			else
				fprintf(f,"0.00000000 ");
		}
		fprintf(f,"\n");
	}
	/*
	if( hist )
		for(int i=0;i<clusterNum;i++)
			fprintf(f,"%f ",hist[i]);
			*/
	//if ( compactness) fprintf(f,"%f",compactness);
	fclose(f);
}

//save histogram hist to txt file
void Corel5k::saveHist(histbin *hist, int imgNum){
	char outFname[255];
	sprintf(outFname,"%s%s\\%d\\%d.txt",baseDir,methodName,imgNum-imgNum%1000,imgNum);
	FILE* f = fopen(outFname,"w");
	int i=0;
	while(hist[i].x!=0)
	{
		fprintf(f,"%4d:%10f ",hist[i].i,hist[i].x);
		if ( i % 5 == 4 )
			fprintf(f,"\n");
		i++;
	}
	fprintf(f,"-1:0");
	fclose(f);
}

//loads descriptor centers from txt file
void Corel5k::loadCenter(int mnr,int imgnum, cv::Mat* centers,int index){
	char fname[256];
	sprintf(fname,"%s%s\\%d\\%d.txt",baseDir,method_names[mnr],imgnum - imgnum%1000,imgnum);
	FILE *f = fopen(fname,"r");

	int i,j;	
	int nrCenters, dim;	
	fscanf(f,"%d%d\n",&nrCenters,&dim);	/*
	if( nrCenters != J || D2 != dim )
	{
		printf("\nError while loading centers, number of centers or dimensions inconsistent!");
		return;
	}*/
	*centers = cv::Mat_<float>(nrCenters, dim);
	//centers = new cv::Mat_<float>(J, D2);
	for(i=0;i<nrCenters;i++)
	{
		for(int d=0;d<dim;d++)
			fscanf(f,"%f",&centers->at<float>(i,d));		
	}	
	for(i=0;i<nrCenters;i++)
		fscanf(f,"%f",hists+index*nrCenters+i);		
	fclose(f);
}

//load sparse histogram
void Corel5k::loadShist(int mnr, int imgnum, int index){
	char fname[256];
	sprintf(fname,"%s%s\\%d\\%d.txt",baseDir,method_names[mnr],imgnum - imgnum%1000,imgnum);
	FILE *f = fopen(fname,"r");

	int i=0; 
	float x=1;
	fscanf(f,"%d:%f",&i,&x);
	while(x!=0){		
		shists[index].push_back(histbin(i,x));
		fscanf(f,"%d:%f",&i,&x);
	}
	fclose(f);
}

//save annotation in txt file by writing word/label indexes only
void Corel5k::saveAnnotation(int imgnum,int labels[5]){
	char outputName[255];
	sprintf(outputName, "%s%s\\test_res%d.txt", baseDir, method_names[0], (int)param.get("ti"));
	FILE *f=fopen(outputName,"a");
	fprintf(f,"\n%d ",imgnum);
	for(int i=0;i<5;i++)
		fprintf(f,"%d ",labels[i]);		
	fclose(f);
};

//creates directories necessary for output
void Corel5k::createDirectoryStructure(){
#ifdef _WIN32
	chdir(baseDir);
	char dirName[255];
	mkdir("__results");
	mkdir(methodName);
	chdir(methodName);
	int prev=0;
	for(int i=0;i<trainImgNr;i++)
		if ( ( imgnums[i] - imgnums[i]%1000 ) != prev )
		{
			prev = imgnums[i] - imgnums[i]%1000;			
			itoa(prev,dirName,10);
			mkdir(dirName);
		}
#endif
}

//calculate feature centers from all training images
void Corel5k::trainAll(){
	strcpy(methodName, method_names[0] );
	createDirectoryStructure();
	currentMethod = MethodRepo::instance()->get(method_names[0]);
	clusterNum = currentMethod->clusterNum;
	dim = currentMethod->dim;

	char fname[255], fname2[256];
	sprintf(fname,"%sdata\\image_nums",baseDir);

	int imgnums[trainImgNr];
	FILE *f = fopen(fname,"r");
	for(int imgnr=0;imgnr<trainImgNr;imgnr++)
		fscanf(f,"%d",imgnums+imgnr);
	fclose(f);
	
	int nr=0,mr;
	int nrKeypoints;
	//float* descr = new float[nrDescr*dim];
	//max nr of descriptors is assumed to be 40k here
	float* descr = new float[40000*dim];
	float* hist = new float[clusterNum];
	double totalTime=0;
	float compactness;
#pragma omp parallel for
	for(int imgnr=0;imgnr<trainImgNr;imgnr++)
	{
		int imgnum = imgnums[imgnr];
		sprintf(fname2,"%s%s\\%d\\%d.txt",baseDir,method_names[0],imgnum - imgnum%1000,imgnum);	
		FILE *ff = 0;
		//if(ff = fopen(fname2,"r") ) { fclose(ff); continue;  }
		//if( imgnr < 2500 ) continue;
		printf("%6.2f%% Training %-7d\n", 100.0*imgnr/trainImgNr, imgnum );
		

		double start=clock();
		sprintf(fname,"%simg\\%d\\%d.jpeg",baseDir,imgnum - imgnum%1000,imgnum);
		IplImage *img = cvLoadImage(fname);				
#if CLUSTERING_TYPE == 0
		applyKmeans(descr, img, currentMethod, hist, fname );
		//showSegmImage(descr, currentMethod->clusterNum, img);
		saveDescr(descr,imgnum,hist);
#elif CLUSTERING_TYPE == 1
		//cvShowImage("original",img);
		//cvMoveWindow("original",0,0);
		applyEm(descr, img, currentMethod, fname2 );		
		
		//applyCvEm(descr, img, currentMethod, fname2 );		
#elif CLUSTERING_TYPE == 2
		applyHierClust(descr, img, currentMethod );
		saveDescr(descr,imgnum);	
#elif CLUSTERING_TYPE == 3
		std::vector<histbin> shist(1);
		findSparseHist(descr, img, currentMethod, &shist );
		saveHist(&shist[0],imgnum);			
#endif
		cvReleaseImage(&img);		
		printf("Time elapsed: %f(sec)\n\n",((double)clock() - start) / CLOCKS_PER_SEC);
		totalTime += ((double)clock() - start) / CLOCKS_PER_SEC;
	};
	delete[] descr;

	sprintf(fname,"%s%s\\TrainingDone.txt",baseDir,methodName);
	f=fopen(fname,"w");
	//int tTime = T(totalTime);
	fprintf(f,"Elapsed time: %d(h):%d(min):%d(sec)\n", (int)totalTime/3600,( (int)totalTime-3600*((int)totalTime/3600)  )/60, (int)totalTime % 60);
	fprintf(f,"Average execution time: %f(sec)\n", totalTime*1.0/trainImgNr);
	fclose(f);

}

//perform annotation on all the test images
void Corel5k::testAll(){
	if( !wh[1] ) readData();
	currentMethod = MethodRepo::instance()->get(method_names[0]);
	clusterNum = currentMethod->clusterNum;
	dim = currentMethod->dim;

	char fname[255];
	/*
	sprintf(fname,"%s%s\\weights.txt",baseDir,currentMethod->name);
	FILE* f = fopen(fname,"r");	
	weights = new float[currentMethod->dim];
	for(int i=0;i<dim;i++)
		fscanf(f,"%f",&weights[i]);
	fclose(f);
	*/

	printf("Loading training data\n");	
	for(int mnr=0;mnr<param.mnr;mnr++)
	{	
		if( centers[mnr*trainImgNr].cols==0)
		for(int i=0;i<trainImgNr;i++)
		//for(int i=0;i<10;i++)
		{
			loadCenter(mnr, imgnums[i], &centers[mnr*trainImgNr+i], i);	
			if(i%10==0)
				printf("\r%s %3.0f%% complete ",method_names[mnr],(i+1)*100.0/trainImgNr);
		}	
	}
	printf("\nDone.\n");
	
	sprintf(fname,"%sdata\\test_1_image_nums",baseDir);
	FILE* f = fopen(fname,"r");	
	int nr=0, imgnum=0;
	float* descr = new float[4000000];
	int labels[5];
	double totalTime=0;
	
	do{
		if( fscanf(f,"%d",&imgnum) < 1 )
			break;
		nr++;
		//if ( nr < 10 ) continue;
		printf("(ti=%3d)%3.0f%% Testing %d\n", (int)param.get("ti"), 100.0*nr/testImgNr , imgnum );
		//param.get("ti")		
		double start=clock();	
		testImgMulti(descr, imgnum, labels, 0);
		//testImgMulti(descr, imgnum, labels, 1);
		printf("Time elapsed: %f(sec)\n\n",((double)clock() - start) / CLOCKS_PER_SEC);
		totalTime += ((double)clock() - start) / CLOCKS_PER_SEC;

	}while(imgnum>0);
	fclose(f);

	delete [] descr;
	char dirName[256];
	sprintf(dirName,"%s%s\\TestDone.txt",baseDir,method_names[0]);
	f=fopen(dirName,"w");
	fprintf(f,"Elapsed time: %d(h):%d(min):%d(sec)\n", (int)totalTime/3600,(int)totalTime/60 - (int)totalTime/3600*60, (int)totalTime % 60);
	fprintf(f,"Average execution time: %f(sec)\n", totalTime*1.0/testImgNr);
	param.setAtr("exec",totalTime*1.0/testImgNr);
	fclose(f);
}

//finds matches once and runs multiple label transfer algorithms for a test image 
void Corel5k::testImgMulti(float* descr, int imgnum, int labels[], int index){
	int nrKeypoints=0;
	char imgName[256];
	sprintf(imgName,"%simg\\%d\\%d.jpeg",baseDir,imgnum - imgnum%1000,imgnum);
	IplImage *img = 0;

	ErrorItem errors[45];	
	for(int mnr=0;mnr<param.mnr;mnr++)
	{

	img = cvLoadImage(imgName);	
	if( !img ){ printf("Cannot open image: %s\n",imgName); return; };	
	currentMethod = MethodRepo::instance()->get(method_names[mnr]);
	clusterNum = currentMethod->clusterNum;
	dim = currentMethod->dim;
	currentMethod->extractFeatures(descr,&nrKeypoints,img);	
	
	float matchErr;
//#pragma omp parallel for// shared(errors)
	for(int i=0;i<trainImgNr;i++)
	//for(int i=0;i<10;i++)
	{
		if (i%100==0)
			printf("\r%3.0f %%",(i+1)*100.0/trainImgNr);
		if( centers[mnr*trainImgNr+i].cols!=0)
		{
			if(!index)
				matchErr = getCompactness(centers[mnr*trainImgNr+i], descr, nrKeypoints, errors[mnr*15+14].err);	
				//matchErr = getCompactnessSSE(centers[mnr*trainImgNr+i], descr, nrKeypoints, 0);	
			else
				matchErr = getHistDist( centers[mnr*trainImgNr+i], descr, hists+i*clusterNum, nrKeypoints);
//#pragma omp critical
			{
				for(int j=0;j<15;j++)
					if( matchErr < errors[mnr*15+j].err )
					{
						for(int k=14;k>j;k--)
						{
						
							errors[mnr*15+k].err = errors[mnr*15+k-1].err;
							errors[mnr*15+k].imgnum = errors[mnr*15+k-1].imgnum;
						}
						errors[mnr*15+j].err = matchErr;
						errors[mnr*15+j].imgnum = imgnums[i];
						break;
					}
			}
			//cvWaitKey();
		}
	}	
	}//end for mnr
	
	
	ErrorItem errors2[45];	
	int ti=(param.get("bag")/50-1)*72;//test index
#if 1
	for(int wf=0;wf<3;wf++)
		for(int knn=5;knn<=15;knn+=5)
			for(int histExp=-2;histExp<=5;histExp++)
#else
	for(int wf=0;wf<1;wf++)
		for(int knn=5;knn<=5;knn+=5)
			for(int histExp=3;histExp<=3;histExp++)
#endif
	{		
	memset(hist,0,classNr*sizeof(float));	
	memcpy(errors2, errors, 45*sizeof(ErrorItem));
		
	float* weights = new float[knn];
	//weigh matches with exponentially decreasing weights
	for(int i=0;i<knn;i++)
		weights[i]=pow(2.0, histExp*(1 - i*1.0/(knn-1)));
#define SHOW_IMAGE 0
#define PRINT 1	
	IplImage* imgMatch=0;
	char matchName[256], title[256];
	if( SHOW_IMAGE )
	{
		imgResize(&img,200);
		cvShowImage("original", img );
	}
	//float *frec;
	float min_err; int dx;
	for(int mnr=0;mnr<param.mnr;mnr++)
	for(int i=0;i<knn;i++)
	{
		if (i==0) dx = 0;	
		ErrorItem ei = errors2[mnr*15+i];
		if(PRINT)printf("\n%-3d : %7f %d ", i, ei.err, ei.imgnum );
		if(i==0)
			min_err = ei.err;

		int j,x;
		for(j=0;j<trainImgNr;j++)
		if( ei.imgnum == imgnums[j])
			break;
		for(int k=0;k<5;k++)
		{
			x = ground_truth[j][k];	
			if ( x < 1 ) break;
			if ( histExp == -1)
				hist[x] += exp( 300*(1 -  ei.err/min_err) ); //works
			else if ( histExp == -2) 
			{
				if(i==0) hist[x] = 10;
				else if ( hist[x]==0) hist[x] = 1;
			}
			else
				hist[x] += weights[i];
		}
				
		if ( SHOW_IMAGE && i < 6){
			if (imgMatch) cvReleaseImage(&imgMatch);									
			//getName("img",matchName,ei.imgnum);
			sprintf(matchName,"%simg\\%d\\%d.jpeg",baseDir,ei.imgnum - ei.imgnum%1000,ei.imgnum);
			imgMatch = cvLoadImage(matchName);	
			if(!imgMatch) { printf("Image not found %s!\n",matchName); }
			else{
			imgResize(&imgMatch,200);
			sprintf(title,"%d match:", i+1);
			cvNamedWindow(title, CV_WINDOW_AUTOSIZE); 			
			cvShowImage(title, imgMatch );
			cvMoveWindow(title ,dx,500);
			dx += imgMatch->width+5;	
			}
		}
	}

	if(wf)
		for(int i=1; i<classNr; i++)
			if( hist[i] )
				hist[i] += wf*0.001*frec[i];
				
	if(PRINT)printf("\nLabels: \n");
	bool mask[classNr];	memset(mask,1,classNr);
	int nr=0, maxi=-1;
	while( nr < 5 )
	{
		maxi = 0;
		for(int i=1;i<classNr;i++)
			if ( hist[i] > hist[maxi] && mask[i])
				maxi = i;
		mask[maxi] = 0;		
		labels[nr] = maxi;
		if(PRINT)
		{
			printf("%s(%d), ",words[labels[nr]],maxi);		
			FILE *ff = fopen("d:\\hist.txt","w");
			for(int i=1;i<classNr;i++)
				if ( hist[i] >= 1 )
					fprintf(ff,"%d\t%s\t%f\n",i,words[i],hist[i]);
			fclose(ff);			
		}
		nr++;		
	}	
	if(PRINT)printf("\n");

	param.setAtr("ti",ti++);	
	saveAnnotation(imgnum,labels);
	if( SHOW_IMAGE) cvWaitKey();
	delete []weights;	
	}
	cvReleaseImage(&img);	
}

//debug
void Corel5k::testImgMulti2(float* descr, int imgnum, int labels[], int index){
	int nrKeypoints=0;
	char imgName[256];
	sprintf(imgName,"%simg\\%d\\%d.jpeg",baseDir,imgnum - imgnum%1000,imgnum);
	IplImage *img = 0;	
	
	std::priority_queue<ErrorItem, std::vector<ErrorItem>, ErrorItemComp> errors[3];
	for(int mnr=0;mnr<param.mnr;mnr++)
	{

	img = cvLoadImage(imgName);	
	if( !img ){ printf("Cannot open image: %s\n",imgName); return; };	
	currentMethod = MethodRepo::instance()->get(method_names[mnr]);
	clusterNum = currentMethod->clusterNum;
	dim = currentMethod->dim;
	currentMethod->extractFeatures(descr,&nrKeypoints,img);	
	
	double matchErr;
//#pragma omp parallel for shared(errors)
	for(int i=0;i<trainImgNr;i++)
	//for(int i=0;i<10;i++)
	{
		if (i%100==0)
			printf("\r%3.0f %%",(i+1)*100.0/trainImgNr);
#if CLUSTERING_TYPE != 1
		if( centers[mnr*trainImgNr+i].cols!=0)
		{
			matchErr = getCompactness(centers[mnr*trainImgNr+i], descr, nrKeypoints, 1e35);	
#else
		if( models[mnr*trainImgNr+i].dim != 0 )
		{
			matchErr = getLogProb( &models[mnr*trainImgNr+i], descr, nrKeypoints, 0 );
#endif
#pragma omp critical
			errors[mnr].push( ErrorItem(matchErr,imgnums[i]) );
			//cvWaitKey();
		}
	}	
	}//end for mnr	

	std::priority_queue<ErrorItem, std::vector<ErrorItem>, ErrorItemComp> errors2[3];
	int ti=(param.get("bag")/50-1)*72;//test index
#if 1
	for(int wf=0;wf<3;wf++)
		for(int knn=5;knn<=15;knn+=5)
			for(int histExp=-2;histExp<=5;histExp++)//*/				
#else
	for(int wf=0;wf<1;wf++)
		for(int knn=5;knn<=5;knn+=5)
			for(int histExp=3;histExp<=3;histExp++)//*/	
#endif
	{		
	memset(hist,0,classNr*sizeof(float));
	std::priority_queue<ErrorItem, std::vector<ErrorItem>, ErrorItemComp> err0(errors[0]);
	std::priority_queue<ErrorItem, std::vector<ErrorItem>, ErrorItemComp> err1(errors[1]);
	std::priority_queue<ErrorItem, std::vector<ErrorItem>, ErrorItemComp> err2(errors[2]);//copy original
	errors2[0] = err0; errors2[1] = err1; errors2[2] = err2;		
		
	float* weights = new float[knn];
	//weigh matches with exponentially decreasing weights
	for(int i=0;i<knn;i++)
		weights[i]=pow(2.0, histExp*(1 - i*1.0/(knn-1)));

#define PRINT 0	
	IplImage* imgMatch=0;
	char matchName[256], title[256];
	if( SHOW_IMAGE)
	{
		imgResize(&img,200);
		cvShowImage("original", img );
	}
	//float *frec;
	float min_err; int dx;
	for(int mnr=0;mnr<param.mnr;mnr++)
	for(int i=0;i<knn;i++)
	{
		if (i==0) dx = 0;	
		ErrorItem ei = errors2[mnr].top();
		if(PRINT)printf("\n%-3d : %7f %d ", i, ei.err, ei.imgnum );
		if(i==0)
			min_err = ei.err;

		int j,x;
		for(j=0;j<trainImgNr;j++)
		if( ei.imgnum == imgnums[j])
			break;
		for(int k=0;k<5;k++)
		{
			x = ground_truth[j][k];	
			if ( x < 1 ) break;
			if ( histExp == -1)
				hist[x] += exp( 300*(1 -  ei.err/min_err) ); //works
			else 
			if ( histExp == -2) 
			{
				if(i==0) hist[x] = 10;
				else if ( hist[x]==0) hist[x] = 1;
			}
			else
				hist[x] += weights[i];	
		}
				
		if ( SHOW_IMAGE && i < 6){
			if (imgMatch) cvReleaseImage(&imgMatch);									
			//getName("img",matchName,ei.imgnum);
			sprintf(matchName,"%simg\\%d\\%d.jpeg",baseDir,ei.imgnum - ei.imgnum%1000,ei.imgnum);
			imgMatch = cvLoadImage(matchName);	
			if(!imgMatch) { printf("Image not found %s!\n",matchName); }
			else{
			imgResize(&imgMatch,200);
			sprintf(title,"%d match:", i+1);
			cvNamedWindow(title, CV_WINDOW_AUTOSIZE); 			
			cvShowImage(title, imgMatch );
			cvMoveWindow(title ,dx,500);
			dx += imgMatch->width+5;	
			}
		}
		errors2[mnr].pop();
	}

	std::priority_queue<ErrorItem, std::vector<ErrorItem> ,ErrorItemComp> labels_queue;
	if(PRINT)printf("\nLabels: \n");
	for(int i=0;i<classNr;i++)
		if(	hist[i]>0 )
		{
			//printf("%d x %s\n",hist[i],words[i]);
			//labels_queue.push( ErrorItem(-hist[i]*1000-param.get("wf")frec[i],i ) );			
			labels_queue.push( ErrorItem((-hist[i])*1000-wf*frec[i],i ) );			
		}
	int nr=0;
	while( nr < 5 && !labels_queue.empty() )
	{
		ErrorItem ei1 = labels_queue.top();
		//strcpy(labels[nr],words[ei1.imgnum]);
		labels[nr] = ei1.imgnum;
		if(PRINT)printf("%s(%d), ",words[labels[nr]],ei1.imgnum);		
		nr++;		
		labels_queue.pop();
	}	
	if(PRINT)printf("\n");

	param.setAtr("ti",ti++);	//todo
	//param.setAtr("ti",0);	
	saveAnnotation(imgnum,labels);
	if( SHOW_IMAGE) cvWaitKey();
	delete []weights;	
	}
	cvReleaseImage(&img);	
}

//calculate precision, recall and nzr
void Corel5k::results(){	
	int wc[classNr],w[classNr];
	//wh - human annotation, is already defined
	if( !wh[1] ) readData();
	for(int i=0;i<classNr;i++)
	{
		wc[i]=0;w[i]=0;
	}
	
	char fname[255];	
	sprintf(fname,"%s%s\\test_res%d.txt",baseDir,method_names[0],(int)param.get("ti"));//todo
	FILE* f = fopen(fname,"r");		
	if( !f ) return;
	int ann[6], j;	
	for(int i=0;i<testImgNr;i++)
	{
		for(int ii=0;ii<6;ii++)
			if ( fscanf(f,"%d",&ann[ii]) < 0 )		
				goto breakfor;					


		//for(int ii=1;ii<1+wordcounts[i];ii++)
		for(int ii=1;ii<6;ii++)
		{
			w[ann[ii]]++;			
			for(j=0;j<5 && test_gt[i][j]>0;j++)
				if (test_gt[i][j] == ann[ii])
				{
					wc[ann[ii]]++;
					break;
				}		

		}			
	}
breakfor:
	fclose(f);

	float prec[classNr],rec[classNr];
	float avgrec=0, avgprec=0;
	int nzr=0, nrwords = 0; //nr of words that appear at least once
	for(int i=1;i<classNr;i++)
	{
		if(wh[i])	
		{
			rec[i] = wc[i]*1.0/wh[i];					
			nrwords++;
		}
		else
			rec[i] = 0;
		if( wc[i] )
			nzr++;

		avgrec += rec[i];
		if(w[i])
		{
			prec[i] = wc[i]*1.0/w[i];			
			//nrwords++;
		}
		else
			prec[i] = 0;
		avgprec += prec[i];
	}	
	avgprec = 100*avgprec/nrwords; 	avgrec = 100*avgrec/nrwords;

	switch(param.mnr)
	{
	default: case 1:
		sprintf(fname,"%s__results\\%s.txt",baseDir,method_names[0]); break;
	case 2:
		sprintf(fname,"%s__results\\%s_%s.txt",baseDir,method_names[0], method_names[1]); break;
	case 3:
		sprintf(fname,"%s__results\\%s_%s_%s.txt",baseDir,method_names[0], method_names[1], method_names[2]);
	}
	f = fopen(fname,"a");
	for(int i=param.trainingIndex;i<param.attrNr;i++)
	{
		fprintf(f, "%s %5.2f ",param.avs[i].attr, param.avs[i].value);
	}	

	//enable this to print individual class precision and recall values
	if(0)
	{
		fprintf(f,"\n");
		for(int i=1;i<classNr;i++)
			if(wc[i]>0)
				fprintf(f, "%4d %15s %7.2f %7.2f\n", i, words[i], 100*prec[i], 100*rec[i]);	
		fprintf(f, "%4d %15s %7.2f %7.2f \n nzr = %4d\n", 0, "average ", avgprec, avgrec, nzr);
	}
	fprintf(f, "%5.2f %5.2f %5.2f %3d\n",
		2*avgprec*avgrec/(avgprec+avgrec), avgprec, avgrec, nzr);
	fclose(f);

}

void Corel5k::cooccurrence(){
	memset(cooc,0,classNr*classNr*sizeof(float));
	for(int i=0;i<trainImgNr;i++)
	{
		for(int j=0;j<wordcounts[i]-1;j++)
			for(int k=j+1;k<wordcounts[i];k++)
			{
				cooc[ground_truth[i][j]*classNr+ground_truth[i][k]]++;
				cooc[ground_truth[i][k]*classNr+ground_truth[i][j]]++;
			}
	}

	/*
	long sumfrec;
	for(int i=1;i<classNr;i++)
	{
		sumfrec = 0;
		for(int j=1;j<classNr;j++)
			sumfrec += cooc[i*classNr+j];	
		if(sumfrec)
		for(int j=1;j<classNr;j++)
			cooc[i*classNr+j] /= sumfrec;
	}*/

	char fname[256];

	sprintf(fname, "%sdata\\cooc", baseDir);
	FILE* f= fopen(fname,"w");
	for(int i=1;i<classNr;i++)
	{
		for(int j=1;j<classNr;j++)
			fprintf(f,"%f ",cooc[i*classNr+j]);
		fprintf(f,"\n");
	}
	fclose(f);
}
