extern "C" {
#include "vl/generic.h"
#include "vl/sift.h"
#include "vl/dsift.h"
}
#include "common.h"
#include "imgOperations.h"
extern params param;

class SIFTMethod : Method{	
public:
	SIFTMethod(){
		clusterNum=20;
		dim=384;
		strcpy(name,"SIFT");
		MethodRepo::instance()->add(name,this);
	}

void _vl_dsift_free_buffers (VlDsiftFilter* self)
{
	if (self->frames) {
		vl_free(self->frames) ;
		self->frames = NULL ;
	}
	if (self->descrs) {
		vl_free(self->descrs) ;
		self->descrs = NULL ;
	}
	if (self->grads) {
		int t ;
		for (t = 0 ; t < self->numGradAlloc ; ++t)
			if (self->grads[t]) vl_free(self->grads[t]) ;
		vl_free(self->grads) ;
		self->grads = NULL ;
	}
	self->numFrameAlloc = 0 ;
	self->numBinAlloc = 0 ;
	self->numGradAlloc = 0 ;
}


void vl_dsift_delete2(VlDsiftFilter *vl_filter)
{
	_vl_dsift_free_buffers (vl_filter) ;
	if (vl_filter->convTmp2) vl_free (vl_filter->convTmp2) ;
	if (vl_filter->convTmp1) vl_free (vl_filter->convTmp1) ;
	vl_free (vl_filter) ;
}

void extractFeatures(float* descr, int* nrKeypoints, IplImage* img){
	//transform to opponent color-space	
#if 1
	IplImage* imgSmall = cvCreateImage(cvGetSize(img),IPL_DEPTH_8U,3);
	cvCopyImage(img, imgSmall);
	imgResize(&imgSmall,320);		
	img = imgSmall;
	IplImage* imgOp =	imgToOCS(img);	
	IplImage* imgO = imgOp;
	IplImage** imgOc = new IplImage*[3];
	for(int c=0;c<img->nChannels;c++)
		imgOc[c] = cvCreateImage(cvGetSize(img),IPL_DEPTH_32F,1);
	for(int c=imgO->nChannels;c<3;c++)
		imgOc[c] = 0;
	cvSplit(imgO,imgOc[0],imgOc[1],imgOc[2],0);
	
#else
	IplImage* imgO = imgToOCS(img);	
	IplImage** imgOc = new IplImage*[3];
	for(int c=0;c<img->nChannels;c++)
		imgOc[c] = cvCreateImage(cvGetSize(img),IPL_DEPTH_32F,1);
	for(int c=imgO->nChannels;c<3;c++)
		imgOc[c] = 0;
	cvSplit(imgO,imgOc[0],imgOc[1],imgOc[2],0);
#endif
		
//extract features in parallel from different channels
#ifdef RUN_HERE
//#pragma omp parallel for shared(descr,nrKeypoints)
#endif
	for(int c=0;c<imgO->nChannels;c++)
	{
		VlDsiftFilter *vl_filter = vl_dsift_new_basic(img->width, img->height, 2, 10);	
		vl_dsift_process(vl_filter, (float*)imgOc[c]->imageData);	
#ifdef RUN_HERE
//#pragma omp critical
#endif
		//if (*nrKeypoints == 0)
		*nrKeypoints = vl_dsift_get_keypoint_num(vl_filter);
		int dsize = vl_dsift_get_descriptor_size(vl_filter);
		float const *descrt = vl_dsift_get_descriptors(vl_filter);

		for (int k=0; k<*nrKeypoints; k++)	
		{
			for (int i=0; i<dsize; i++)
			{
				if ( isnan(descrt[k*dsize+i]) )
					descr[k*imgO->nChannels*dsize+c*dsize+i]=0.0;
				else
					descr[k*imgO->nChannels*dsize+c*dsize+i]=descrt[k*dsize+i];
				//if ( descr[(k*dsize)+i] < 0)
				//	printf("Warning negative descriptor found!\n");
			}				
			//descrt += dsize;			
		}
		vl_dsift_delete2(vl_filter);
	}
	for(int c=0;c<imgO->nChannels;c++)
		cvReleaseImage(&imgOc[c]);
	cvReleaseImage(&imgO);
	cvReleaseImage(&imgSmall);
	delete [] imgOc;	
}

}_sift;