#include "common.h"
#include "imgoperations.h"
#include "Gabor.h"

extern params param;

//extract simple local color descriptor
class ColorDescr:Method{
public:
	ColorDescr(){
		clusterNum = 20;
		dim = 9;
		strcpy(name,"color");
		MethodRepo::instance()->add(name,this);
	}

	void extractFeatures(float* descr, int* nrKeypoints, IplImage* img){

		IplImage* imgSmall = cvCreateImage(cvGetSize(img),IPL_DEPTH_8U,3);
		cvCopyImage(img, imgSmall);
		imgResize(&imgSmall,64);	
		IplImage* imgSmallLab = imgToLab(imgSmall);
		IplImage* imgSmallHSV = imgToHSV(imgSmall);
		//IplImage* imgSmallHSV = imgToOCS(imgSmall);

		*nrKeypoints=imgSmall->width*imgSmall->height;
		uchar* data = (uchar*) imgSmall->imageData;
		uchar* dataLab = (uchar*) imgSmallLab->imageData;
		uchar* dataHSV = (uchar*) imgSmallHSV->imageData;
		//float* dataHSV = (float*) imgSmallHSV->imageData; // HSV works better

		for(int i=0; i<*nrKeypoints; i++)
		{
			descr[i*dim+0] = data[i*3];
			descr[i*dim+1] = data[i*3+1];
			descr[i*dim+2] = data[i*3+2];
			//descr[i*dim+3] = i / img->width;
			//descr[i*dim+4] = (i % img->width);

			//*/			
			
			descr[i*dim+3] = dataLab[i*3];
			descr[i*dim+4] = dataLab[i*3+1];
			descr[i*dim+5] = dataLab[i*3+2];
			//*/

			descr[i*dim+6] = dataHSV[i*3];
			descr[i*dim+7] = dataHSV[i*3+1];
			descr[i*dim+8] = dataHSV[i*3+2];
			//*/
		}	
		cvReleaseImage(&imgSmall);cvReleaseImage(&imgSmallLab);cvReleaseImage(&imgSmallHSV);

		#if 1
			FILE* f = fopen("d:/descr.txt","w");
			for(int i=0;i<*nrKeypoints;i++)
			{
				for(int j=0;j<dim;j++)
					fprintf(f,"%f\t",descr[i*dim+j]);
				fprintf(f,"\n");
			}
			fclose(f);
		#endif
	}
}_color;

class ColorOcs:Method{//23%
public:
	ColorOcs(){
		clusterNum = 20;
		dim = 3;
		strcpy(name,"color-ocs");
		MethodRepo::instance()->add(name,this);
	}

	void extractFeatures(float* descr, int* nrKeypoints, IplImage* img){	
		IplImage* imgSmall = cvCreateImage(cvGetSize(img),IPL_DEPTH_8U,3);
		cvCopyImage(img, imgSmall);
		imgResize(&imgSmall,64);	
		IplImage* imgSmallOcs = imgToOCS(imgSmall);

		*nrKeypoints=imgSmall->width*imgSmall->height;
		float* data = (float*) imgSmallOcs->imageData;
		memcpy( descr, data, *nrKeypoints*3*sizeof(float));
		/*
		for(int i=0; i<*nrKeypoints; i++)
		{
			descr[i*dim+0] = data[i*3];
			descr[i*dim+1] = data[i*3+1];
			descr[i*dim+2] = data[i*3+2];			
		}*/	
		cvReleaseImage(&imgSmall);cvReleaseImage(&imgSmallOcs);
	}
}_colorOcs;


class Gabor: public Method{
	int nrbins;
public:
	Gabor(){
		clusterNum = 20;
		dim = 9;
		strcpy(name,"Gabor");
		MethodRepo::instance()->add(name,this);
		nrbins = 16;
	}
void extractFeatures(float* descr, int* nrKeypoints, IplImage* img){
	imgResize(&img,120);
	IplImage** imgGabor = new IplImage*[12];
	IplImage* imgG = imgToGray(img);
	char title[256];
	int nr=0;
	for(int angle=0;angle<8;angle+=2)
		for(int scale=1;scale<4;scale++)
		{
		imgGabor[nr] = cvCreateImage(cvGetSize(img),IPL_DEPTH_8U, 1);
		CvGabor filter(angle,(int)pow(2.0,(scale-1)*1.0),1);
		filter.conv_img(imgG,imgGabor[nr],CV_GABOR_MAG);		
		/*
		pos = ( angle/2*3 + (scale-1) )*nrbins;
		for(int i=0;i<nrbins;i++)
			(X+pos)[i] = 0;
		for(int i=0;i<imgOut->width*imgOut->height;i++)
			if ( (int)data[i]/nrbins > -1 && (int)data[i]/nrbins < nrbins)
				(X+pos)[ (int)data[i]/nrbins ]++;
			else
				printf("Value out of range at %d = %d",i,(int)data[i]/nrbins);
		for(int i=0;i<nrbins;i++)
			(X+pos)[i] /= img->width*img->height;*/
		if (SHOW_IMAGE){
		sprintf(title,"angle=%d, scale=%d",(int)(angle*22.5),(int)pow(2.0,(scale-1)*1.0));
		cvShowImage(title,imgGabor[nr]);
		cvMoveWindow(title,200*(nr/3),200*(nr%3));
		}
		nr++;
		}
		if (SHOW_IMAGE) cvWaitKey();
		//optional histogram calculation on nxn block

		if (nr != dim )	return;
		*nrKeypoints = (imgGabor[0]->width-1)*(imgGabor[0]->height-1);
		for(int i=0;i<*nrKeypoints;i++)
			for(int j=0;j<nr;j++)
			{
				descr[i*nr+j] = ((uchar*)imgGabor[j]->imageData)[i];
			}

		cvReleaseImage(&imgG);
		for(int j=0;j<nr;j++)
			cvReleaseImage(&imgGabor[j]);
		cvReleaseImage(&img);	
}
};

//*/
class WLDdescr:Method{
public:
	int angle_bins, mag_bins, nh_size;

	WLDdescr(){
		clusterNum = 20;
		angle_bins = 8;
		mag_bins = 6;
		nh_size = 8;
		dim = angle_bins*mag_bins;
		strcpy(name,"WLD");
		MethodRepo::instance()->add(name,this);
	}

void extractFeatures(float* X, int* nrKeypoints, IplImage* img){
	//extracts local descriptors as a WLD histograms of a 8x8 patches
	//calculates differential excitation and gradient orientation for every image pixel
	//		
	//float weights[mag_bins] = {0.2688, 0.0852, 0.0955, 0.1000, 0.1018, 0.3487}; //what about weights???
	IplImage* imgG = imgToGray(img);	
	IplImage* imgDiffEx = cvCreateImage(cvGetSize(imgG),IPL_DEPTH_32F,1);
	IplImage* imgDeg = cvCreateImage(cvGetSize(imgG),IPL_DEPTH_32F,1);

	uchar* data = (uchar*) imgG->imageData;
	int w = imgG->widthStep;
	int w2 = imgDeg->width; //not widthstep becasue float* handles data size

	cvZero(imgDiffEx);cvZero(imgDeg);
	float dx,dy,de;
	int r = 1;
	for(int i=r;i<imgG->height-r;i++){
		for(int j=r;j<imgG->width-r;j++){			
			dx = data[i*w+j+1] - data[i*w+j-1];
			dy = data[(i+1)*w+j] - data[(i-1)*w+j];			
			de = 0;
			for(int ii=-r;ii<=r;ii++)
				for(int jj=-r;jj<=r;jj++)
					if(  data[i*w+j] && (abs(ii) == r || abs(jj) == r ))
					{
						de += (data[(i+ii)*w+j+jj] - data[i*w+j])*1.0 / data[i*w+j];
					}					
			((float*)imgDiffEx->imageData)[i*w2+j] = atan(de);
			((float*)imgDeg->imageData)[i*w2+j] = atan2(dy, dx)+PI;
		}
	}
	float* magnitudes = (float*)imgDiffEx->imageData;
	float* degrees = (float*)imgDeg->imageData;
	int delta=2, nr=0;
	const float dt = 2*PI/angle_bins;
	memset(X,0,imgG->height*imgG->width*dim/delta/delta*sizeof(float));
	for(int i=1;i<imgG->height-nh_size;i+=delta){
		for(int j=1;j<imgG->width-nh_size;j+=delta)
		{			
			for(int ii=0;ii<nh_size;ii++)
				for(int jj=0;jj<nh_size;jj++)
				{	
					int t = ((int)(degrees[(i+ii)*w2+ j+jj] / dt + 0.5)) % angle_bins;
					int m = (int)( (magnitudes[(i+ii)*w2+ j+jj]+PI/2)*mag_bins/PI );	
					X[nr*angle_bins*mag_bins+m*mag_bins+t]++; 
					//descr[nr*angle_bins*mag_bins+m*mag_bins+t] += weights[m];
				}
			nr++;
		}
	}	
	*nrKeypoints = nr;
	cvReleaseImage(&imgG);cvReleaseImage(&imgDiffEx);cvReleaseImage(&imgDeg);
}
}_wld;

class ColorMoments:Method{
public:

	ColorMoments(){
		clusterNum = 20;
		dim = 21;
		strcpy(name,"colormoments");
		MethodRepo::instance()->add(name,this);
	}

float moment( uchar* data, int type ) {
	type++;
	float res = 1.f; int div = 0;
	if ( type & 1 )
	{
		res *= data[0];		
		div++;
	}
	if ( type & 2 )
	{
		res *= data[1];		
		div++;
	}
	if ( type & 4 )
	{
		res *= data[2];
		div++;
	}
	if ( div == 1)
		return res;//?
	else
		if (res<0)
			return -pow( (float) -res, 1.0f/div );
		else
			return pow( (float) res, 1.0f/div );
	}

void extractFeatures(float* X, int* nrKeypoints, IplImage* img){
	const int d = 4;
	const int width = img->width;
	const int height = img->height;
	int* xs = new int[d*2+1];
	for(int i=0;i<d*2+1;i++)
		xs[i] = i-d;

	uchar* data = (uchar*)img->imageData;
	*nrKeypoints = 0;
	memset(X, 0, width*height*dim/4*sizeof(float));
	for(int i=d+1; i<height - d - 1; i+=2)
		for(int j=d+1; j<width - d - 1; j+=2)
		{
			for(int k=0;k<dim;k++)
				for(int x=-d; x<=d; x++)
					for(int y=-d; y<=d; y++)
						if ( k < 7 )
						{
							X[(*nrKeypoints)*dim + k] += moment( data+(i+x)*3*width + (j+y)*3, k );
						}
						else if ( k < 14 )
						{
							X[(*nrKeypoints)*dim + k] += xs[x+d]*moment( data+(i+x)*3*width + (j+y)*3, k-7 );
						}
						else
						{
							X[(*nrKeypoints)*dim + k] += xs[y+d]*moment( data+(i+x)*3*width + (j+y)*3, k-14 );
						}
			(*nrKeypoints)++;
		}

	const float inv = 1.0/((2*d+1)*(2*d+1));
	for(int i=0;i<(*nrKeypoints)*dim;i++)
		X[i] *= inv;
	}	
}_cm;

//Law energy mask for texture description
void conv(IplImage* img, IplImage* dst,int kernel[5][5],int size){
	//IplImage* dst = cvCreateImage(cvGetSize(img), IPL_DEPTH_32F, 1);
	cvSetZero(dst);
	int d=size/2;
	
	float tmp;
	int w= img->width;
	uchar* data = (uchar*)img->imageData;
	for(int i=d;i<img->height-d;i++)
		for(int j=d;j<img->width-d;j++)
		{	
			tmp=0;
			for(int ii=-d;ii<=d;ii++)
				for(int jj=-d;jj<=d;jj++)		
					tmp += data[(i+ii)*w + j+jj] * kernel[ii+d][jj+d];
			((float*)dst->imageData)[i*w+j]=tmp;
		}
}

/*
void extractLaw(float* X, int* nrKeypoints, IplImage* img){
	//source http://www.ccs3.lanl.gov/~kelly/ZTRANSITION/notebook/laws.shtml
	//if (lawKernels[0][0][0] != 1)		generateLawKernels();
	IplImage* imgG = imgToGray(img);
	
	//generateLawKernels();
	int dd=7, bs=2*dd+1;
	int w=imgG->width;
	IplImage* imgLaw[16];
	IplImage* imgEnLaw[16];
	float*  test1 ,*test2;
	for(int k=0;k<16;k++)
	{		
		//imgLaw[k] = conv(imgG, lawKernels[k], 5);
		//  test1 =  (float* ) imgLaw[k]->imageData;
		imgEnLaw[k] = cvCreateImage(cvGetSize(imgG), IPL_DEPTH_32F, 1);
		cvSetZero(imgEnLaw[k]);
		for(int i=dd;i<imgG->height-dd;i+=2)
			for(int j=dd;j<imgG->width-dd;j+=2)
			{				
				for(int ii=-dd;ii<=dd;ii++)
					for(int jj=-dd;jj<=dd;jj++)				
						((float*)imgEnLaw[k]->imageData)[i*w+j] += abs( ((float*)imgLaw[k]->imageData)[(i+ii)*w + j+jj] ) ;
			}
	//	 test2 = (float* ) imgEnLaw[k]->imageData;
	}
	
	int nr=0,kk=0;
	int dim=param.get("totalDim");
	float tmp;
	for(int i=dd;i<imgG->height-dd;i+=2)
		for(int j=dd;j<imgG->width-dd;j+=2)
		{
			tmp = ((float*)imgEnLaw[0]->imageData)[i*w+j];
			kk=0;
			for(int k1=1;k1<4;k1++)
				for(int k2=0;k2<=k1;k2++)
				{
					if ( k1 == k2 )
						X[nr*dim+kk] = ((float*)imgEnLaw[4*k1+k2]->imageData)[i*w+j] / tmp;
					else if (tmp)
						X[nr*dim+kk] = (  ((float*)imgEnLaw[4*k1+k2]->imageData)[i*w+j] + ((float*)imgEnLaw[4*k2+k1]->imageData)[i*w+j]  )
											/ 2 / tmp;
					else
						X[nr*dim+kk] = 0;
					kk++;
				}
			nr++;			
		}
		*nrKeypoints = nr;
	
	for(int k=0;k<16;k++)
	{
		cvReleaseImage(&imgLaw[k]);cvReleaseImage(&imgEnLaw[k]);
	}
	cvReleaseImage(&imgG);

}
//*/

class LawDescr:Method{
public:
	//the 15 kernels, each 5x5
	int lawKernels[16][5][5];

	LawDescr(){
		clusterNum = 20;
		dim = 30;//4*5/2 per channel
		strcpy(name,"Law-color");
		MethodRepo::instance()->add(name,this);
	}

void inline outerProduct(int res[5][5], int* x, int* y){
	for(int i=0;i<5;i++)
		for(int j=0;j<5;j++)
			res[i][j] = x[i]*y[j];
	}

void generateLawKernels(){
		int K[4][5] = {
		 { 1, 4, 6, 4, 1}, //L
		 {-1,-2, 0, 2, 1}, //E
		 {-1, 0, 2, 0,-1}, //S
		// {-1, 2, 0,-2, 1}, //W
		 { 1,-4, 6,-4, 1}  //R
		};	
		
		for(int i=0;i<4;i++)
			for(int j=0;j<4;j++)
				outerProduct(lawKernels[i*4+j],K[i],K[j]);
	}

void extractFeatures(float* X, int* nrKeypoints, IplImage* img){
	//source http://www.ccs3.lanl.gov/~kelly/ZTRANSITION/notebook/laws.shtml
	//here the three channels of the image is convolved with 16 law kernels 
	//the descriptors will be these values (no energy calculation)
	if (lawKernels[0][0][0] != 1)
		generateLawKernels();
	IplImage* imgs[3];
	for(int c=0;c<3;c++) imgs[c] = cvCreateImage(cvGetSize(img),IPL_DEPTH_8U,1);
	cvSplit(img,imgs[0],imgs[1],imgs[2],0);
	
	IplImage* imgLaw[48];
	for(int i=0;i<48;i++) imgLaw[i] = cvCreateImage(cvGetSize(img), IPL_DEPTH_32F, 1);
//#pragma omp parallel for
	for(int k=0;k<16;k++)
		for(int c=0;c<3;c++)		
		{		
			conv(imgs[c], imgLaw[c*16+k], lawKernels[k], 5);
		}

	*nrKeypoints=0;
	int nr,i1,i2;
	int w=img->width;
	for(int i=5;i<img->height-5;i+=2)
		for(int j=5;j<w-5;j+=2)
		{
			nr = 0;
			for(int c=0;c<3;c++)
			{
				//normalize with 25  (the size of the kernel)
				X[nr++] = (((float*)imgLaw[c*16]->imageData)[i*w+j] + ((float*)imgLaw[c*16]->imageData)[i*w+j])*1.0/25;						
				for(int k1=1;k1<4;k1++)
					for(int k2=0;k2<=k1;k2++)
					{
						i1 = c*16 + k1*4 + k2; i2 = c*16 + k2*4 + k1;
						X[nr++] = ((float*)imgLaw[i1]->imageData)[i*w+j] + ((float*)imgLaw[i2]->imageData)[i*w+j];						
					}
			}
			X += dim;
			(*nrKeypoints)++;
		}	
	
	for(int c=0;c<3;c++)
		cvReleaseImage(&imgs[c]);
	for(int k=0;k<48;k++)	
		cvReleaseImage(&imgLaw[k]);	
	//cvReleaseImage(&img);
}
}_law;


