#include "common.h"
#include "imgoperations.h"

void imgResize(IplImage **img, int maxDim){
	//resizes all images (shrink or enlarge) so that the bigger dimension is equal to maxDim (320)
	//maxdim is 320 by default
	int height    = (*img)->height;
	int width     = (*img)->width;
	int channels  = (*img)->nChannels;
	if ( height != maxDim && width!= maxDim && (height > maxDim || width > maxDim))
	{	
		if ( !height*width )
			return;
		
		float ratio = maxDim*1.0f / max(height, width); 
		IplImage *dst;
		if ( height != maxDim )		
			dst = cvCreateImage(cvSize(maxDim, (int)( ratio* height )), IPL_DEPTH_8U, channels);
		else
			dst = cvCreateImage(cvSize((int)( ratio* width ), maxDim), IPL_DEPTH_8U, channels);

		cvResize( *img, dst );
		cvReleaseImage(img);
		*img = dst;
	}
}

void imgResize(IplImage *img, IplImage** dst, int maxDim ){
	//resizes all images (shrink or enlarge) so that the bigger dimension is equal to maxDim (320)
	//maxdim is 320 by default
	int height    = img->height;
	int width     = img->width;
	int channels  = img->nChannels;
	if ( height != maxDim && width!= maxDim && (height > maxDim || width > maxDim))
	{	
		if ( !height*width )
			return;
		
		float ratio = maxDim*1.0f / max(height, width); 
		//IplImage *dst;
		if ( height != maxDim )		
			*dst = cvCreateImage(cvSize(maxDim, (int)( ratio* height )), IPL_DEPTH_8U, channels);
		else
			*dst = cvCreateImage(cvSize((int)( ratio* width ), maxDim), IPL_DEPTH_8U, channels);

		cvResize( img, *dst );
	}
}

//create output and convert img to grayscale
IplImage* imgToGray(IplImage *img){
	IplImage *imgOut = cvCreateImage(cvGetSize(img),IPL_DEPTH_8U,1);
	cvCvtColor(img, imgOut, CV_BGR2GRAY);		
	return imgOut;
}

//create output and convert img to Lab colorspace
IplImage* imgToLab(IplImage *img){
	IplImage *imgOut = cvCreateImage(cvGetSize(img),IPL_DEPTH_8U,3);
	cvCvtColor(img, imgOut, CV_BGR2Lab);		
	return imgOut;
}

//create output and convert img to HSV colorspace
IplImage* imgToHSV(IplImage *img){
	IplImage *imgOut = cvCreateImage(cvGetSize(img),IPL_DEPTH_8U,3);
	cvCvtColor(img, imgOut, CV_BGR2HSV);		
	return imgOut;
}

//create output and convert img to opponent-color colorspace
float isqrt2 = 1.0f/sqrt(2.0f);
float isqrt3 = 1.0f/sqrt(3.0f);
float isqrt6 = isqrt2*isqrt3;
IplImage* imgToOCS(IplImage* img){
	IplImage *imgOut = cvCreateImage(cvGetSize(img),IPL_DEPTH_32F,3);
	uchar* datain = (uchar*) img->imageData;
	float* dataout = (float*) imgOut->imageData;

	uchar r,g,b;
	for(int i=0;i<img->height;i++)
		for(int j=0;j<img->width;j++)
		{
			r = datain[i*3*img->width+j*3+2];
			g = datain[i*3*img->width+j*3+1];
			b = datain[i*3*img->width+j*3  ];

			dataout[i*3*img->width+j*3  ]=(r-g)*isqrt2;		//O1
			dataout[i*3*img->width+j*3+1]=(r+g-2*b)*isqrt6;	//O2
			dataout[i*3*img->width+j*3+2]=(r+g+b)*isqrt3;	//O3
		}
	return imgOut;
}

//convert to opponent-color space and put results in imgOut
void imgToOCS(IplImage* img, IplImage* &imgOut){
	imgOut = cvCreateImage(cvGetSize(img),IPL_DEPTH_32F,3);
	uchar* datain = (uchar*) img->imageData;
	float* dataout = (float*) (imgOut)->imageData;

	uchar r,g,b;
	for(int i=0;i<img->height;i++)
		for(int j=0;j<img->width;j++)
		{
			r = datain[i*3*img->width+j*3+2];
			g = datain[i*3*img->width+j*3+1];
			b = datain[i*3*img->width+j*3  ];

			dataout[i*3*img->width+j*3  ]=(r-g)*isqrt2;		//O1
			dataout[i*3*img->width+j*3+1]=(r+g-2*b)*isqrt6;	//O2
			dataout[i*3*img->width+j*3+2]=(r+g+b)*isqrt3;	//O3
		}
}

//create output and convert to normalised opponent color-space, 2 channels
IplImage* imgToNOCS(IplImage* img){
	IplImage *imgOut = cvCreateImage(cvGetSize(img),IPL_DEPTH_32F,2);
	uchar* datain = (uchar*) img->imageData;
	float* dataout = (float*) imgOut->imageData;

	float isqrt2 = 1.0f/sqrt(2.0f);
	float sqrt3 = sqrt(3.0f);	

	uchar r,g,b;
	for(int i=0;i<img->height;i++)
		for(int j=0;j<img->width;j++)
		{
			r = datain[i*3*img->width+j*3+2];
			g = datain[i*3*img->width+j*3+1];
			b = datain[i*3*img->width+j*3  ];

			dataout[i*2*img->width+j*2  ]=(r-g)*sqrt3/(r+g+b);		//O1/O3
			dataout[i*2*img->width+j*2+1]=(r+g-2*b)*isqrt2/(r+g+b);	//O2/O3			
		}
	return imgOut;
}

