#include "common.h"

//1 - absolute dist, 2 - Euclidian, 3 chi-square
#define DIST_TYPE 1

template <class T> inline T distX( T x, T y){
#if DIST_TYPE == 1
	return fabs(x-y);
#elif DIST_TYPE == 2
	return (x-y)*(x-y);
#elif DIST_TYPE == 3
	if ( x + y )
		return (x-y)*(x-y)/(x+y);
	else
		return 0;
#endif
}

//applies k-means on the data X, TT is the number of instances, K is the number of clusters, D is dimension
//X contains the data in a row-major order (first D values are the first instance etc.)
//centers will contain the cluster centers in a row-major order
float kmeansAbs(float* centers, float* X, int _TT, int _K, int _D)
{
	const int K = _K;
	const int D = _D;
	const int TT = _TT;
	char* labels = new char[TT];
	int* nrelems  = new int[K];
	float* tmpcenters = new float[K*D];

#if DIST_TYPE == 1
	vector< vector< vector<float> > > elems(D);
	for(int i=0;i<D;i++)
	{
		elems[i].resize(K);
		for(int k=0;k<K;k++)
			elems[i][k].resize(TT-K+1);//presumption - no cluster can contain more than this nr
	}
#endif

	float minl = 1e50;
	for(int seed=1; seed < 1+5; seed++)
	{
	srand(seed);
		
#define KMEANSPP
#ifdef KMEANSPP
	{
	//Arthur Vassilvitskii initialization
	// runs in O(K*TT*D)
	int k=0;
	int randi = rand() %TT;
	memcpy( tmpcenters+(k++)*D, X+randi*D, D*sizeof(float) );
	vector< vector<float> > dists(K);
	float* cumdists = new float[TT]; //for probability distr
	float dist, mind;
	do{
		//find min-dists
		dists[k-1].resize(TT);
		for(int t=0;t<TT;t++)
		{
			mind = 1e30;
			dist = 0;
			for(int d=0;d<D;d++)
				dist += distX( tmpcenters[(k-1)*D+d], X[t*D+d] );
			dists[(k-1)][t] = dist;

			for(int kk=0;kk<k;kk++)
				if ( dists[kk][t] < mind )
					mind = dists[kk][t];

			if( t == 0 )
				cumdists[t] = mind;//*mind;
			else
				cumdists[t] = cumdists[t-1] + mind;//*mind;
		}

		//sample next center 
		float randf = rand()*cumdists[TT-1] / RAND_MAX;
		int tt;
		for(tt=0;tt<TT && cumdists[tt] < randf;tt++);
		memcpy( tmpcenters+(k++)*D, X+tt*D, D*sizeof(float) );

	}while(k<K);
	delete[] cumdists;
	}
#else	
	//randomly select first centers - O(K)
	int randi;
	for(int k=0;k<K;k++)
	{
		randi = rand() %TT;
		memcpy( tmpcenters+k*D, X+randi*D, D*sizeof(float) );
	}	
#endif

	float dist, mind, diff; int lbl;
	//const int mindiff = max((int)(0.005 * TT), 5); ?? maybe stopping shouldn't be based on an absolute value
	int iter;
	for(iter=0;iter<100;iter++){
		
		//find labels, i.e. classify data, associate with centroids
		diff = 0;
		for(int t=0;t<TT;t++)
		{
			mind=1e30;
			for(int k=0;k<K;k++)
			{
				dist = 0;
				for(int d=0;d<D;d++)
					dist += distX( tmpcenters[k*D+d], X[t*D+d] );
				if( dist < mind )
				{
					mind = dist;
					lbl = k;
				}
			}
			if ( lbl != labels[t] )
				diff++;
			labels[t] = lbl;
		}
		//if ( diff < 5 ) break; //if no. of reassignments is small stop the alg.
		if ( diff < 0.005*TT) break;
		//printf("iter: %d\n",iter);
		
#if DIST_TYPE > 0
		//recalculate centroids
		memset(tmpcenters, 0, K*D*sizeof(float));
		memset(nrelems, 0, K*sizeof(int));
		for(int t=0;t<TT;t++)
		{
			nrelems[labels[t]]++;
			for(int d=0;d<D;d++)
				tmpcenters[labels[t]*D+d] += X[t*D+d];
		}
		for(int k=0;k<K;k++)
			if ( nrelems[k] )
				for(int d=0;d<D;d++)
					tmpcenters[k*D+d] /= (nrelems[k]);
			else
			{
				// in this case this center has no elements
				// add another random center and continue
				printf("should not be here\n");
				//int randi = rand() %TT;
				//memcpy( tmpcenters+k*D, X+randi*D, D*sizeof(float) );
			}
#else
		//in this case take the median
		// dE/dm_i = #{ x_i > m_i } -  #{ x_i < m_i } = 0 => m_i is the median of x_i-s
		memset(nrelems, 0, K*sizeof(int));		
		for(int t=0;t<TT;t++)
		{
			for(int d=0;d<D;d++)
				elems[d][labels[t]][nrelems[labels[t]]] = X[t*D+d];
			nrelems[labels[t]]++;
		}
		for(int d=0;d<D;d++)
			for(int k=0;k<K;k++)
				sort(elems[d][k].begin(), elems[d][k].begin()+nrelems[k]);

		for(int d=0;d<D;d++)
			for(int k=0;k<K;k++)
				tmpcenters[k*D+d] = elems[d][k][nrelems[k]/2];

#endif

#if 0
		//calculate the compactness at each iteration
	{
	double ll=0, ll2, tmp, minn;
	for(int t=0;t<TT;t++)
	{
		minn = 1e30;
		for(int k=0;k<K;k++)
		{
			tmp = 0;
			for(int d=0;d<D;d++)
				tmp += distX( tmpcenters[k*D+d], X[t*D+d] );
			if ( tmp < minn )
				minn = tmp;
		}
		ll += minn;
	}
	printf("\tCompactness (%d iters) = %f\n",iter,ll);
	}
#endif
	}//end iters
	
	//calculate objective function only once at the end
	double ll=0, ll2, tmp, minn;
	for(int t=0;t<TT;t++)
	{
		minn = 1e30;
		for(int k=0;k<K;k++)
		{
			tmp = 0;
			for(int d=0;d<D;d++)
				tmp += distX( tmpcenters[k*D+d], X[t*D+d] );
			if ( tmp < minn )
				minn = tmp;
		}
		ll += minn;
	}
	if(0) printf("\tCompactness (%d iters) = %f\n",iter,ll);
	if ( ll < minl )
	{
		minl = ll;			
		memcpy( centers, tmpcenters, K*D*sizeof(float));
		//printf("*\n");
	}
	//else
		//printf("\n");

	}//end seed for
	
	delete[] labels; delete [] nrelems; delete[] tmpcenters;
	return minl;
}