#include "common.h"
#include "kmeansBased.h"
#include "kmeans.h"
#include "imgOperations.h"
//#include "kpp/kmeans.h"

extern params param;
extern cv::Mat* centers;
extern int* imgnums;
extern int** ground_truth;
extern Method *currentMethod;

using namespace cv;

//testing function
void extractKmeansDCT(float* X, IplImage* img){
	int nrKeypoints=0;
	IplImage* imgLab = imgToLab(img);
//	DCTvector(X, imgLab, &nrKeypoints, 2);
	cv::Mat points, centers, clusters;	
	points = (cv::Mat_<float>(nrKeypoints, (int)param.get("nrDimDct")));
	points.data = (uchar*) X;//why uchar?

	int clusterNum = (int)param.get("clusterNum");
	double compactness = cv::kmeans( points, clusterNum, clusters, cvTermCriteria(CV_TERMCRIT_ITER+CV_TERMCRIT_EPS, 200, 1e-7 ), 5, cv::KMEANS_PP_CENTERS, centers);	
	centers.release(); points.release(); clusters.release();
}

float inline distL1(float x, float y){
	//return abs(x-y);	
	float ff = x-y;
	int* i = (int*)&ff;
	*i &= 0x7fffffff;
	return (*(float*)i);	
	//return x>y ? x-y : y-x;
}
float inline distL2(float x, float y){
	return (x-y)*(x-y);
}
float inline distChi(float x, float y){
	if (x+y)	return (x-y)*(x-y)/(x+y);
	else		return 0;
}

/*
returns how close the data in X is to the centers 
by evaluating the mean distance from the closest
centers
*/
double getCompactness(cv::Mat centers, float* X,int nrKeypoints,float maxComp){
	
	//int dt = param.get("delta");
	int dt = (int)nrKeypoints/param.get("bag");
	//int dt=10;	
	const int clusterNum = centers.rows;
	const int totalDim = centers.cols;
	/*
	float (*distX)(float, float)=0;
	switch((int)param.get("metric"))
	{
		default:
		case 1:	distX = &distL1; break;
		case 2:	distX = &distL2; break;
		case 3:	distX = &distChi; break;		 
	}
	//*/
	double compactness=0,dist,mindist;
	int i,j,t;
	for(t=0;t<nrKeypoints;t=t+dt)
	{
		mindist=DBL_MAX;
		for(i=0;i<clusterNum;i++)
		//	if ( centers.at<float>(i,0) > 0)
		{
			dist = 0;
			for(j=0;j<totalDim;j++)
			{				
				//dist += distX(X[j],centers.at<float>(i,j));				
				//printf("X[%d]=%f\n
				dist += distL1(X[j],centers.at<float>(i,j));				
				//dist += weights[j]*distL1(X[j],centers.at<float>(i,j));				
				if (dist > mindist)
				//	break;
				goto breaker;
			}			
			//if (dist<mindist)
			mindist = dist;			
			//compactness += dist;
breaker:
				dist=dist;
		}
		X+=dt*totalDim;	
		compactness += mindist;
		if ( compactness > maxComp )
			return compactness;
	}

	//return compactness/nrKeypoints;
	return compactness;//param.get("bag");
}

void findHist(float* centers, float* descr, float* hist, int nrKeypoints, int clusterNum, int dim){
	memset(hist,0,clusterNum*4);
	int bag = (int)param.get("bag");
	int dt;
	if ( bag > 0 )
		dt = nrKeypoints/bag;
	else
		dt = 1;
	for(int i=0;i<nrKeypoints;i+=dt)
	{
		int minj = -1;
		float mind = 1e10;
		float dist = 0;
		for(int j=0;j<clusterNum;j++)
		{
			dist = 0;
			for(int k=0;k<dim && dist<mind;k++)
				dist += fabs( descr[k] - centers[j*dim+k] );
			if ( dist < mind )
			{
				mind = dist;
				minj = j;
			}
		}
		hist[minj] += mind;
		//hist[minj]++;
		descr += dt*dim;
	}
#if 0
	/*
	for(int i=0;i<clusterNum;i++)
		hist[i] /= nrKeypoints;
	*/
#else
	//L1 normalization - local bag of words
	float sum=0;
	for(int i=0;i<clusterNum;i++)
		sum += hist[i];
	for(int i=0;i<clusterNum;i++)
		hist[i] /= sum;
#endif
}

float getHistDist(cv::Mat centers, float* X, float* hist, int nrKeypoints){
	float *histTest = new float[centers.rows];
	findHist((float*)(centers.datastart), X, histTest, nrKeypoints, centers.rows, centers.cols );
	float dist = 0 ;
	for(int i=0;i<centers.rows;i++)
		dist += fabs( hist[i] - histTest[i] );
		//dist += histTest[i];
	delete[] histTest;
	return dist;
}

// Sparse histogram stuff: ////////////////////////////////////////////////////////////////
void h_increment(histbin* hist, int& len, TTT pos, int dim){
	if (!len)
	{
		hist[len++].set(pos,1);
		hist[len].i = 100000;
		return;
	}	
	
	int start=0; int end=len; int middle;
	while( end - start > 1 ){
		middle = (end+start)/2;
		if ( hist[middle].i < pos )
			start = middle;
		else
			end = middle;
	}

	if ( hist[end].i == pos )
		hist[end].x++;
	else if ( hist[start].i == pos )
		hist[start].x++;
	else
	{
		if ( pos < hist[start].i ) 
			end = start;		
		len++;
		for(int j=len;j>end;j--)
			hist[j] = hist[j-1];
		hist[end].set(pos,1);
	}
	
}

int getSparseHist(float* X, int nrKeypoints, int dim, histbin* hist, float minx, float maxx, int k){

	float* limits = new float[k];
	float dl = minx+(maxx - minx)/k;
	limits[0] = dl;
	for(int kk=1;kk<k;kk++)
		limits[kk] = limits[kk-1]+dl;

	memset(hist, 0, nrKeypoints*sizeof(histbin));
	int len=0;
	for(int i=0;i<nrKeypoints;i++)
	{
		int pos=0;
		//vect pos(dim);
		for(int d=0;d<dim;d++)
		{
			int kk=0;
			for(kk=0;kk<k;kk++)
				if (X[d] < limits[kk])
					break;
			
			pos += kk;	
			if ( d != dim-1)
				pos *= k;
				
			//pos.x[d] = (TTT)kk;
		}		
			
		h_increment(hist, len, pos, dim);
		X += dim;
	}
#if 0
	for(int i=0; i<nrKeypoints; i++)
		hist[i].x /= nrKeypoints;
#endif
	delete [] limits;
	return len;
}

void findSparseHist(float* centers, IplImage* img, Method* mm, std::vector<histbin>* hist){
	int clusterNum = mm->clusterNum;
	int dim = mm->dim;
	int nrKeypoints=0;	
	float* descr = new float[dim*26000]; // 160x160<26k
	
	mm->extractFeatures(descr,&nrKeypoints,img);	
	histbin* hist_tmp = new histbin[nrKeypoints];
	int len = getSparseHist(descr,nrKeypoints,dim,hist_tmp,0,256,16);	
	hist->resize(len+1);
	for(int i=0;i<=len;i++)
		(*hist)[i] = hist_tmp[i];
	delete[] hist_tmp;
}

float getHistDist(histbin* h, int lenh, histbin* g, int leng){
	int i=0,j=0;
	float dist=0;
	do{
		if ( h[i].i < g[j].i )
		{
			dist += h[i++].x;
		}
		else if ( h[i].i > g[j].i )
		{
			dist += g[j++].x;
		}
		else 
			dist += fabs( h[i++].x - g[j++].x );
	}while( i < lenh && j < leng );

	if ( j < lenh )
		do{
			dist += g[j++].x;
		}while( j < leng );
	else 
		if ( i < lenh )
		do{
			dist += h[i++].x;
		}while( i < lenh );

	return dist;
}

void loadFeatures(float* descr, int* nrKeypoints, char* fname){
	FILE* f = fopen(fname, "r");
	int dim;
	fscanf(f,"%d%d",nrKeypoints,&dim);
	bool iszero = true;
	int zeros = 0;
	for(int i=0;i<*nrKeypoints;i++)
	{
		iszero = true;
		for(int j=0;j<dim;j++)
		{
			fscanf(f,"%f",descr+j);
			if ( descr[j] ) 
				iszero = false;
		}
		if ( !iszero )
			descr += dim;
		else
			zeros++;
	}
	fclose(f);
	*nrKeypoints -= zeros;
}


// wrapper function for applying k-means on the features from img
void applyKmeans(float* centers, IplImage* img, Method* mm, float* hist, char* fname){
	int clusterNum = mm->clusterNum;
	int dim = mm->dim;
	int nrKeypoints=0;	
	float* descr = new float[dim*26000]; // 160x160<26k

	//loadFeatures(descr,&nrKeypoints,fname);
	mm->extractFeatures(descr,&nrKeypoints,img);

#define TRIALS 5
#define KTYPE 0
#if KTYPE == 0
	kmeansAbs(centers, descr, nrKeypoints, clusterNum, dim);
#elif KTYPE == 1
	int * assignments = new int[nrKeypoints];
	float comp = RunKMeansPlusPlus(nrKeypoints, clusterNum, dim, descr, TRIALS, centers, assignments);
	printf("\tCompactness = %f\n",comp);
	delete[] assignments;
#elif KTYPE == 2
	int * assignments = new int[nrKeypoints];
	float comp = RunKMeans(nrKeypoints, clusterNum, dim, descr, TRIALS, centers, assignments);
	printf("\tCompactness = %f\n",comp);
	delete[] assignments;
#elif KTYPE == 3	
	//opencv kmeans
	//corelSaveDescr(descr,1000);
	cv::Mat points, cvcenters, clusters;	
	points = (cv::Mat_<float>(nrKeypoints, dim));	
	points.data = (uchar*) descr;
	
	double compactness = cv::kmeans( points, clusterNum, clusters, cvTermCriteria(CV_TERMCRIT_ITER+CV_TERMCRIT_EPS, 200, 1e-7 ), TRIALS, cv::KMEANS_PP_CENTERS, cvcenters);	
	for(int i=0;i<dim*clusterNum;i++)
		centers[i] = ((float*)cvcenters.data)[i];
	cvcenters.release(); 	points.release(); clusters.release();
	printf("\tCompactness = %f\n",compactness);
#endif

#if 0
	//create a histogram of distances to the closest centers
	findHist(centers,descr,hist,nrKeypoints,clusterNum,dim);	
#endif

	delete[] descr;
}